"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Optional, Union

from arelle.ModelValue import qname, QName, DateTime, YearMonthDayTimeDuration
from arelle.PythonUtil import OrderedSet
from .XbrlConst import xbrl
from .XbrlProperty import XbrlProperty
from .XbrlTypes import XbrlTaxonomyType, QNameKeyType, DefaultTrue, DefaultFalse
from .ModelValueMore import QNameAt, SQName
from .XbrlTaxonomyObject import XbrlTaxonomyObject, XbrlReferencableTaxonomyObject

class XbrlDateResolution(XbrlTaxonomyObject):
    conceptName: Optional[QName] # (optional) Identifies the QName of a concept object that has a date fact value. The values of the concept object resolves to a set of dates. If no value exists in the report then the property is ignored, and no date constraint is enforced on the cube.
    context: Optional[QNameAt] # (optional) Identifies the QName of a concept object that has a value. The context of the fact values resolves to a set of dates. If no value exists in the report then the property is ignored. The context suffix must be either @end or @start. If an @ value is not provided then the suffix defaults to @end.
    value: Optional[DateTime] # (optional) A literal date value representing the end date.
    timeShift: Optional[YearMonthDayTimeDuration] # (optional) Defines a time duration shift from the date derived form either the value, context or name properties. The duration of the time shift is defined using the XML duration type to define a duration of time. A negative operator is used to deduct the timeShift from the resolved date.

class XbrlPeriodConstraint(XbrlTaxonomyObject):
    periodType: str # (required) Used to indicate if the period is an instant or a duration.
    timeSpan: Optional[str] # (optional) Defines a duration of time using the XML duration type to define a duration of time. The duration of the time span maps to facts with the same duration.
    periodFormat: Optional[str] # (optional) Defines a a duration of time with an end date. The period value defined in the taxonomy must resolve to a valid period format as defined in xbrl-csv specification.
    monthDay: Optional[XbrlDateResolution] # (optional) Represents a Gregorian date that recurs such as 04-16. The date resolution object when used with this property returns the date without the year. The conceptName property of the dateResolution object can also use a datatype of monthDay.
    endDate: Optional[XbrlDateResolution] # (optional) Defines an end date for a duration fact and the date of an instant fact. Values can be provided as a literal date value, a fact with a date value, or the date context value of a date. A suffix of @start or @end may be added to any of the date formats, specifying the instant at the start or end end of the duration, respectively.
    startDate: Optional[XbrlDateResolution] # (optional) Defines a start date for a duration fact and the date of an instant fact. Values can be provided as a literal date value, a fact with a date value, or the date context value of a date. A suffix of @start or @end may be added to any of the date formats, specifying the instant at the start or end end of the duration, respectively.
    onOrAfter: Optional[XbrlDateResolution] # (optional) Defines a date where all instant facts on or after the date are included in the cube. For a duration fact any periods after or on the end date of the duration are included in the cube.
    onOrBefore: Optional[XbrlDateResolution] # (optional) Defines a date where all instant facts before or on the date are included in the cube. For a duration fact any periods before or on the end date are included in the cube.

class XbrlCubeDimension(XbrlTaxonomyObject):
    dimensionName: QName # (required) The QName of the dimension object that is used to identify the dimension. For the core dimensions of concept, period, entity and unit, the core dimension QNames of xbrl:concept, xbrl:period, xbrl:entity, xbrl:unit and xbrl:language are used. The dimension object indicates if the dimension is typed or explicit.
    domainName: Optional[QName] # (required if explicit dimension) The QName of the domain object that is used to identify the domain associated with the dimension. Only one domain can be associated with a dimension. The domain name cannot be provided for a typed dimension or the period core dimension.
    domainSort: Optional[str] # (optional if typed dimension) A string value that indicates the sort order of the typed dimension. The values can be either asc or desc. The values are case insensitive. This indicates if the cube is viewed the order of the values shown on the typed dimension. This cannot be used on an explicit dimension.
    allowDomainFacts: Union[bool, DefaultFalse] # (optional    ) A boolean value that indicates if facts not identified with the dimension are included in the cube. For typed and explicit dimensions the value defaults to false. A value of true for a typed or explicit dimension will include facts that don't use the dimension in the cube. For the period core dimension, forever facts or facts with no period dimension are included when this value is set to true. For units, this is a unit with no units such as a string or date. For the entity core dimension, it is fact values with no entity. This property cannot be used on the concept core dimension.
    periodConstraints: set[XbrlPeriodConstraint] # (optional only for period core dimension) Defines an ordered set of periodConstraint objects to restrict fact values in a cube to fact values with a specified period.

class XbrlCube(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: QNameKeyType # (required) The name property is a QName that uniquely identifies the cube object.
    cubeType: Optional[QName] # (optional) The cubeType property identifies the type of data cube being represented. This must match a defined cubeType object or specification defined cube types of xbrl:eventCube, xbrl:positionCube, xbrl:referenceCube, xbrl:reportCube, xbrl:journalCube, xbrl:eventDetailsCube, xbrl:timeSeriesCube and xbrl:defaultCube. If no QName is provided the default is xbrl:reportCube.
    cubeDimensions: OrderedSet[XbrlCubeDimension] # (required) An ordered set of cubeDimension objects that identify the dimensions and associated domains used on the cube.
    cubeNetworks: OrderedSet[QName] # (optional) An ordered set of network object QNames that reference network objects that are directly related to the cube.
    excludeCubes: OrderedSet[QName] # (optional) An ordered set of cube object QNames that remove the facts of the constraint cube from the facts of the defined cube.
    cubeComplete: Optional[bool] # (optional) A boolean flag that indicates if all cells in the cube are required to have a value. If true then all cube cells must include a fact value. If a value is not provided for the cubeComplete property then the default is false.
    properties: OrderedSet[XbrlProperty] # (optional) An ordered set of property objects Used to specify additional properties associated with the cube using the property object. Only immutable properties as defined in the propertyType object can be added to a cube.

class XbrlAllowedCubeDimension(XbrlTaxonomyObject):
    dimensionName: Optional[QName] # (optional) The dimension QName that identifies the taxonomy defined dimension.
    dimensionType: Optional[str] # (optional) The dimension QName that identifies the taxonomy defined dimension.
    dimensionDataType: Optional[QName] # (optional) The dimension QName that identifies the taxonomy defined dimension.
    required: Union[bool, DefaultFalse] # (optional) The dimension QName that identifies the taxonomy defined dimension.

class XbrlRequiredCubeRelationship(XbrlTaxonomyObject):
    relationshipTypeName: QName # (required) The relationship type QName of a relationship. This requires that at lease one of these relationship types exist on the cube.
    source: Optional[QName] # (optional) The QName of the source object type in the relationship.
    target: Optional[QName] # (optional) The QName of the target object type in the relationship.

class XbrlCubeType(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: QNameKeyType # (required) The name is a QName that uniquely identifies the cube type object.
    baseCubeType: Union[bool, DefaultTrue] # (optional) Base cube type that the cube object is based on. Uses the QName of a cubeType object. The property only allows restriction rather than expansion of the baseCubeTape.
    periodDimension: Union[bool, DefaultTrue] # (optional) boolean to indicate if the period core dimension is included in the cube. Defaults to true.
    entityDimension: Union[bool, DefaultTrue] # (optional) boolean to indicate if the entity core dimension is included in the cube. Defaults to true.
    unitDimension: Union[bool, DefaultTrue] # (optional) boolean to indicate if the unit core dimension is included in the cube. Defaults to true.
    taxonomyDefinedDimension: Union[bool, DefaultTrue] # (optional) boolean to indicate if taxonomy defined dimensions are included in the cube. Defaults to true.
    allowedCubeDimensions: OrderedSet[XbrlAllowedCubeDimension] # (optional) An ordered set of allowedCubeDimension objects that are permitted to be used on the cube. If the property is not defined then any dimensions can be associated with the cube.
    requiredCubeRelationships: OrderedSet[XbrlRequiredCubeRelationship] # (optional) An ordered set of requiredCubeRelationship objects that at a minimum must be associated with the cube.

baseCubeTypes = {
    qname(xbrl, "xbrl:eventCube"),
    qname(xbrl, "xbrl:positionCube"),
    qname(xbrl, "xbrl:referenceCube"),
    qname(xbrl, "xbrl:reportCube"),
    qname(xbrl, "xbrl:journalCube"),
    qname(xbrl, "xbrl:eventDetailsCube"),
    qname(xbrl, "xbrl:timeSeriesCube"),
    qname(xbrl, "xbrl:defaultCube")
    }

periodCoreDim = qname(xbrl, "xbrl:period")
conceptCoreDim = qname(xbrl, "xbrl:concept")
entityCoreDim = qname(xbrl, "xbrl:entity")
unitCoreDim = qname(xbrl, "xbrl:unit")
languageCoreDim = qname(xbrl, "xbrl:language")

coreDimensions = {periodCoreDim, conceptCoreDim, entityCoreDim, unitCoreDim, languageCoreDim}

conceptDomainRoot = qname(xbrl, "xbrl:conceptDomain")
entityDomainRoot = qname(xbrl, "xbrl:entityDomain")
unitDomainRoot = qname(xbrl, "xbrl:unitDomain")



"""
See COPYRIGHT.md for copyright information.

The `arelle.api` module is the supported method for integrating Arelle into other Python applications.
"""
from __future__ import annotations

import logging
import threading
from types import TracebackType
from typing import Any, BinaryIO, TypeVar

from arelle import PackageManager, PluginManager
from arelle.CntlrCmdLine import CntlrCmdLine, createCntlrAndPreloadPlugins
from arelle.FileSource import FileNamedBytesIO
from arelle.ModelXbrl import ModelXbrl
from arelle.RuntimeOptions import RuntimeOptions

_session_lock = threading.Lock()

# typing.Self can be used once Python 3.10 support is dropped.
Self = TypeVar("Self", bound="Session")


class Session:
    """
    CRITICAL THREAD SAFETY WARNING:

    Arelle uses shared global state (PackageManager, PluginManager) which is NOT thread-safe.
    Only ONE Session can run at a time across the entire process.

    Safe usage:
    - Use one Session at a time per process
    - Use a process pool instead of thread pool for parallelism

    Unsafe usage:
    - Running multiple Sessions concurrently in any threads
    - Threading.Thread with Session.run()
    """

    def __init__(self) -> None:
        self._cntlr: CntlrCmdLine | None = None
        self._thread_id = threading.get_ident()

    def _check_thread(self) -> None:
        """Ensure session is only used from the thread that created it."""
        if threading.get_ident() != self._thread_id:
            raise RuntimeError(
                "Session objects cannot be shared between threads. Create a new Session instance in each thread."
            )

    def __enter__(self: Self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType | None
    ) -> None:
        self.close()

    def close(self) -> None:
        with _session_lock:
            self._check_thread()
            if self._cntlr is not None:
                self._cntlr.close()
            PluginManager.close()

    def get_log_messages(self) -> list[dict[str, Any]]:
        """
        :return: Raw log records (messages) from the session.
        """
        if not self._cntlr or not self._cntlr.logHandler:
            return []
        return getattr(self._cntlr.logHandler, 'messages', [])

    def get_logs(self, log_format: str, clear_logs: bool = False) -> str:
        """
        Retrieve logs as a string in the configured format.
        Raises a `ValueError` if the log format is not supported by the current log handler.
        Optionally clear the log buffer after retrieving.
        :param log_format: The format to retrieve logs in. Supported formats are 'json', 'text', and 'xml'.
        :param clear_logs: If enabled, clears the log buffer after retrieving logs.
        :return:
        """
        if self._cntlr:
            handler = self._cntlr.logHandler
            if log_format == 'json' and hasattr(handler, 'getJson'):
                return str(handler.getJson(clearLogBuffer=clear_logs))
            if log_format == 'text' and hasattr(handler, 'getText'):
                return str(handler.getText(clearLogBuffer=clear_logs))
            if log_format == 'xml' and hasattr(handler, 'getXml'):
                return str(handler.getXml(clearLogBuffer=clear_logs, includeDeclaration=False))
            raise ValueError('Unsupported log format for {}: {}'.format(type(handler).__name__, log_format))
        return ""

    def get_models(self) -> list[ModelXbrl]:
        """
        Retrieve a list of loaded models.
        """
        if self._cntlr is None:
            return []
        return self._cntlr.modelManager.loadedModelXbrls

    def run(
        self,
        options: RuntimeOptions,
        sourceZipStream: BinaryIO | FileNamedBytesIO | None = None,
        responseZipStream: BinaryIO | None = None,
        logHandler: logging.Handler | None = None,
        logFilters: list[logging.Filter] | None = None,
    ) -> bool:
        """
        Perform a run using the given options.
        When using a sourceZipStream, the name attribute on it will be checked
        and, if present, assumed to be the effective name of the zip file the
        stream represents. When validating '.xbri' files, via a sourceZipStream,
        it is important to set this name attribute to avoid
        rpe:documentTypeFileExtensionMismatch errors.

        :param options: Options to use for the run.
        :param sourceZipStream: Optional stream to read source data from.
        :param responseZipStream: Options stream to write response data to.
        :param logHandler: Optional log handler to use for logging.
        :return: True if the run was successful, False otherwise.
        """
        with _session_lock:
            self._check_thread()
            PackageManager.reset()
            PluginManager.reset()
            if self._cntlr is None:
                # Certain options must be passed into the controller constructor to have the intended effect
                self._cntlr = createCntlrAndPreloadPlugins(
                    uiLang=options.uiLang,
                    disablePersistentConfig=options.disablePersistentConfig,
                    arellePluginModules={},
                )
            else:
                # Certain options passed into the controller constructor need to be updated
                if self._cntlr.uiLang != options.uiLang:
                    self._cntlr.setUiLanguage(options.uiLang)
                self._cntlr.disablePersistentConfig = options.disablePersistentConfig or False
            logRefObjectProperties = True
            if options.logRefObjectProperties is not None:
                logRefObjectProperties = options.logRefObjectProperties
            if options.webserver:
                assert sourceZipStream is None, "Source streaming is not supported with webserver"
                assert responseZipStream is None, "Response streaming is not supported with webserver"
                if not self._cntlr.logger:
                    self._cntlr.startLogging(
                        logFileName='logToBuffer',
                        logFilters=logFilters,
                        logHandler=logHandler,
                        logTextMaxLength=options.logTextMaxLength,
                        logRefObjectProperties=logRefObjectProperties,
                        logPropagate=options.logPropagate,
                    )
                    self._cntlr.postLoggingInit()
                from arelle import CntlrWebMain
                CntlrWebMain.startWebserver(self._cntlr, options)
                return True
            else:
                if not self._cntlr.logger:
                    self._cntlr.startLogging(
                        logFileName=(options.logFile or "logToPrint"),
                        logFileMode=options.logFileMode,
                        logFormat=(options.logFormat or "[%(messageCode)s] %(message)s - %(file)s"),
                        logLevel=(options.logLevel or "DEBUG"),
                        logFilters=logFilters,
                        logHandler=logHandler,
                        logToBuffer=options.logFile == 'logToBuffer',
                        logTextMaxLength=options.logTextMaxLength,  # e.g., used by EDGAR/render to require buffered logging
                        logRefObjectProperties=logRefObjectProperties,
                        logXmlMaxAttributeLength=options.logXmlMaxAttributeLength,
                        logPropagate=options.logPropagate,
                    )
                    self._cntlr.postLoggingInit()  # Cntlr options after logging is started
                return self._cntlr.run(
                    options,
                    sourceZipStream=sourceZipStream,
                    responseZipStream=responseZipStream,
                )



"""
See COPYRIGHT.md for copyright information.

## Overview

The Save Loadable OIM plugin facilitates the saving of re-loadable instances in JSON or CSV formats, adhering to the
Open Information Model (OIM) XBRL Specification. It is designed to work seamlessly with the Load From OIM plugin,
allowing for efficient data handling in Arelle.

## Key Features

- **Multiple Formats**: Enables saving data in JSON and CSV OIM formats as well as XLSX.
- **Seamless Integration**: Compatible with the Load From OIM plugin for saving and loading reports.
- **GUI and CLI Compatibility**: Available for use in both GUI and CLI modes.
- **Test Case Augmentation**: Allows saving OIM files with a custom suffix during test suite runs.

## Usage Instructions

### Command Line Usage

- **Single-Instance Mode**:
  Save a file in single-instance mode by specifying the file path and extension:
  ```bash
  python arelleCmdLine.py --plugins saveLoadableOIM --file filing-documents.zip --saveLoadableOIM example.json
  ```

- **Test Case Operation**:
  Augment test case operations by specifying a suffix for the read-me-first file in a test suite:
  ```bash
  python arelleCmdLine.py --plugins saveLoadableOIM --file filing-documents.zip --saveTestcaseOimFileSuffix -savedOim.csv
  ```

- **Deduplicate facts**
  To save an OIM instance with duplicate fact removed use the `--deduplicateOimFacts` argument with either `complete`,
  `consistent-pairs`, or `consistent-sets` as the value.
  For details on what eaxctly consitutes a duplicate fact and why there are multiple options read the
  [Fact Deduplication][fact-deduplication] documentation.
  ```bash
  python arelleCmdLine.py --plugins saveLoadableOIM --file filing-documents.zip --saveLoadableOIM example.json --deduplicateOimFacts complete
  ```

[fact-deduplication]: project:/user_guides/fact_deduplication.md

### GUI Usage

- **Save Re-Loadable Output**:
  1. Load the desired report in Arelle.
  2. Go to `Tools` > `Save Loadable OIM`.
  3. Specify a filename and choose the desired file format (JSON, CSV, or XLSX).

## Additional Notes

If the loaded report refers to a taxonomy on the local file system, the OIM instance needs to be saved in the same
directory to maintain the validity of the relative file system path to the taxonomy.
"""

from __future__ import annotations

import csv
import io
import json
import operator
import os
import threading
import zipfile
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from math import isinf, isnan
from numbers import Number
from optparse import OptionParser
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, Callable, Optional, cast

import regex as re
from openpyxl import Workbook
from openpyxl.cell.cell import WriteOnlyCell
from openpyxl.styles import Alignment, Color, PatternFill, fills
from openpyxl.worksheet.dimensions import ColumnDimension

from arelle import ValidateDuplicateFacts, XbrlConst
from arelle.ModelDocumentType import ModelDocumentType
from arelle.ModelInstanceObject import ModelContext, ModelFact
from arelle.ModelRelationshipSet import ModelRelationshipSet
from arelle.ModelValue import (
    DateTime,
    DayTimeDuration,
    IsoDuration,
    QName,
    Time,
    YearMonthDuration,
    gDay,
    gMonth,
    gMonthDay,
    gYear,
    gYearMonth,
    qname,
    tzinfoStr,
)
from arelle.typing import TypeGetText
from arelle.UrlUtil import relativeUri
from arelle.utils.PluginData import PluginData
from arelle.utils.PluginHooks import PluginHooks
from arelle.ValidateXbrlCalcs import inferredDecimals
from arelle.Version import authorLabel, copyrightLabel

if TYPE_CHECKING:
    from tkinter import Menu

    from arelle.Cntlr import Cntlr
    from arelle.CntlrCmdLine import CntlrCmdLine
    from arelle.CntlrWinMain import CntlrWinMain
    from arelle.ModelXbrl import ModelXbrl
    from arelle.RuntimeOptions import RuntimeOptions

_: TypeGetText

PLUGIN_NAME = "Save Loadable OIM"

oimErrorPattern = re.compile("oime|oimce|xbrlje|xbrlce")
nsOim = "https://xbrl.org/2021"
qnOimConceptAspect = qname("concept", noPrefixIsNoNamespace=True)
qnOimLangAspect = qname("language", noPrefixIsNoNamespace=True)
qnOimPeriodAspect = qname("period", noPrefixIsNoNamespace=True)
qnOimEntityAspect = qname("entity", noPrefixIsNoNamespace=True)
qnOimUnitAspect = qname("unit", noPrefixIsNoNamespace=True)

reservedUriAliases = {
    nsOim: "xbrl",
    XbrlConst.defaultLinkRole: "_",
    XbrlConst.factExplanatoryFact: "explanatoryFact",
    XbrlConst.factFootnote: "footnote",
    XbrlConst.iso4217: "iso4217",
    XbrlConst.utr: "utr",
    XbrlConst.xbrli: "xbrli",
    XbrlConst.xsd: "xs",
}

ONE = Decimal(1)
TEN = Decimal(10)
NILVALUE = "nil"
SCHEMA_LB_REFS = {
    qname("{http://www.xbrl.org/2003/linkbase}schemaRef"),
    # qname("{http://www.xbrl.org/2003/linkbase}linkbaseRef")
}
ROLE_REFS = {
    qname("{http://www.xbrl.org/2003/linkbase}roleRef"),
    qname("{http://www.xbrl.org/2003/linkbase}arcroleRef"),
}
ENTITY_NA_QNAME = ("https://xbrl.org/entities", "NA")
csvOpenMode = "w"
csvOpenNewline = ""

OimFact = dict[str, Any]
OimReport = dict[str, Any]

class NamespacePrefixes:
    def __init__(self, prefixesByNamespace: dict[str, str] | None = None) -> None:
        self._prefixesByNamespace: dict[str, str] = prefixesByNamespace or {}
        self._usedPrefixes: set[str] = set(self._prefixesByNamespace.values())

    @property
    def namespaces(self) -> dict[str, str]:
        return {
            prefix: namespace
            for namespace, prefix in sorted(
                self._prefixesByNamespace.items(),
                key=operator.itemgetter(1)
            )
        }

    def __contains__(self, namespace: str) -> bool:
        return namespace in self._prefixesByNamespace

    def getPrefix(self, namespace: str) -> str | None:
        return self._prefixesByNamespace.get(namespace)

    def addNamespace(self, namespace: str, preferredPrefix: str) -> str:
        prefix = self._prefixesByNamespace.get(namespace)
        if prefix is not None:
            return prefix

        prefix = reservedUriAliases.get(namespace)
        if prefix is None:
            prefix = preferredPrefix
            i = 2
            while prefix in self._usedPrefixes:
                prefix = f"{preferredPrefix}{i}"
                i += 1
        self._prefixesByNamespace[namespace] = prefix
        self._usedPrefixes.add(prefix)
        return prefix


def saveLoadableOIM(
    modelXbrl: ModelXbrl,
    oimFile: str,
    outputZip: zipfile.ZipFile | None = None,
    # arguments to add extension features to OIM document
    extensionPrefixes: dict[str, str] | None = None,
    extensionReportObjects: dict[str, Any] | None = None,
    extensionFactPropertiesMethod: Callable[[ModelFact, OimFact], None] | None = None,
    extensionReportFinalizeMethod: Callable[[OimReport], None] | None = None,
    *args: Any,
    **kwargs: Any,
) -> None:
    isJSON = oimFile.endswith(".json")
    isCSV = oimFile.endswith(".csv")
    isXL = oimFile.endswith(".xlsx")
    isCSVorXL = isCSV or isXL
    if not isJSON and not isCSVorXL:
        oimFile = oimFile + ".json"
        isJSON = True

    namespacePrefixes = NamespacePrefixes({nsOim: "xbrl"})
    if extensionPrefixes:
        for extensionPrefix, extensionNamespace in extensionPrefixes.items():
            namespacePrefixes.addNamespace(extensionNamespace, extensionPrefix)
    linkTypeAliases = {}
    groupAliases = {}

    def compileQname(qname: QName) -> None:
        if qname.namespaceURI is not None and qname.namespaceURI not in namespacePrefixes:
            namespacePrefixes.addNamespace(qname.namespaceURI, qname.prefix or "")

    aspectsDefined = {qnOimConceptAspect, qnOimEntityAspect, qnOimPeriodAspect}

    def oimValue(obj: Any) -> Any:
        if isinstance(obj, list):
            # set-valued enumeration fact
            return " ".join([oimValue(o) for o in obj])
        if isinstance(obj, QName) and obj.namespaceURI is not None:
            if obj.namespaceURI not in namespacePrefixes:
                namespacePrefixes.addNamespace(obj.namespaceURI, obj.prefix or "_")
            return f"{namespacePrefixes.getPrefix(obj.namespaceURI)}:{obj.localName}"
        if isinstance(obj, (float, Decimal)):
            try:
                if isinf(obj):
                    return "-INF" if obj < 0 else "INF"
                elif isnan(obj):
                    return "NaN"
                elif isinstance(obj, Decimal):
                    # XML canonical representation of decimal requires a decimal point.
                    # https://www.w3.org/TR/xmlschema-2/#decimal-canonical-representation
                    if obj % 1 == 0:
                        return f"{obj:.1f}"
                    intPart, fracPart = f"{obj:f}".split(".")
                    canonicalFracPart = fracPart.rstrip("0") or "0"
                    return f"{intPart}.{canonicalFracPart}"
                else:
                    return f"{obj}"
            except Exception:
                return str(obj)
        if isinstance(obj, bool):
            return "true" if obj else "false"
        if isinstance(
            obj,
            (
                DateTime,
                YearMonthDuration,
                DayTimeDuration,
                Time,
                gYearMonth,
                gMonthDay,
                gYear,
                gMonth,
                gDay,
                IsoDuration,
                int,
            ),
        ):
            return str(obj)
        return obj

    def oimPeriodValue(cntx: ModelContext) -> dict[str, str]:
        if cntx.isStartEndPeriod:
            s = cntx.startDatetime
            e = cntx.endDatetime
        else:  # instant
            s = e = cntx.instantDatetime
        assert isinstance(s, datetime) and isinstance(e, datetime)
        return {
            str(qnOimPeriodAspect): (
                f"{s.year:04}-{s.month:02}-{s.day:02}T{s.hour:02}:{s.minute:02}:{s.second:02}{tzinfoStr(s)}/"
                if cntx.isStartEndPeriod
                else ""
            )
            + f"{e.year:04}-{e.month:02}-{e.day:02}T{e.hour:02}:{e.minute:02}:{e.second:02}{tzinfoStr(e)}"
        }

    hasTuple = False
    hasLang = False
    hasUnits = False
    hasNumeric = False

    footnotesRelationshipSet = ModelRelationshipSet(modelXbrl, "XBRL-footnotes")

    # compile QNames in instance for OIM
    for fact in modelXbrl.factsInInstance:
        concept = fact.concept
        if concept is not None:
            if concept.isNumeric:
                hasNumeric = True
            if concept.baseXbrliType in ("string", "normalizedString", "token") and fact.xmlLang:
                hasLang = True
        compileQname(fact.qname)
        if hasattr(fact, "xValue") and isinstance(fact.xValue, QName):
            compileQname(fact.xValue)
        unit = fact.unit
        if unit is not None:
            hasUnits = True
        if fact.modelTupleFacts:
            hasTuple = True
    if hasTuple:
        modelXbrl.error(
            "arelleOIMsaver:tuplesNotAllowed", "Tuples are not allowed in an OIM document", modelObject=modelXbrl
        )
        return

    entitySchemePrefixes = {}
    for cntx in modelXbrl.contexts.values():
        if cntx.entityIdentifierElement is not None:
            scheme = cntx.entityIdentifier[0]
            if scheme not in entitySchemePrefixes:
                if not entitySchemePrefixes:  # first one is just scheme
                    if scheme == "http://www.sec.gov/CIK":
                        _schemePrefix = "cik"
                    elif scheme == "http://standard.iso.org/iso/17442":
                        _schemePrefix = "lei"
                    else:
                        _schemePrefix = "scheme"
                else:
                    _schemePrefix = f"scheme{len(entitySchemePrefixes) + 1}"
                namespacePrefixes.addNamespace(scheme, _schemePrefix)
                entitySchemePrefixes[scheme] = namespacePrefixes.getPrefix(scheme)
        for dim in cntx.qnameDims.values():
            compileQname(dim.dimensionQname)
            aspectsDefined.add(dim.dimensionQname)
            if dim.isExplicit:
                compileQname(dim.memberQname)

    for unit in modelXbrl.units.values():
        if unit is not None:
            for measures in unit.measures:
                for measure in measures:
                    compileQname(measure)

    if hasLang:
        aspectsDefined.add(qnOimLangAspect)
    if hasUnits:
        aspectsDefined.add(qnOimUnitAspect)

    for footnoteRel in footnotesRelationshipSet.modelRelationships:
        if footnoteRel.arcrole not in linkTypeAliases:
            typePrefix = reservedUriAliases.get(footnoteRel.arcrole, f"ftTyp_{os.path.basename(footnoteRel.arcrole)}")
            linkTypeAliases[footnoteRel.arcrole] = typePrefix
        if footnoteRel.linkrole not in groupAliases:
            groupPrefix = reservedUriAliases.get(footnoteRel.linkrole, f"ftGrp_{os.path.basename(footnoteRel.linkrole)}")
            groupAliases[footnoteRel.linkrole] = groupPrefix

    dtsReferences = set()
    assert modelXbrl.modelDocument is not None
    baseUrl = modelXbrl.modelDocument.uri.partition("#")[0]
    for doc, ref in sorted(modelXbrl.modelDocument.referencesDocument.items(), key=lambda _item: _item[0].uri):
        if ref.referringModelObject.qname in SCHEMA_LB_REFS:
            dtsReferences.add(relativeUri(baseUrl, doc.uri))
    for refType in ("role", "arcrole"):
        for refElt in sorted(
            modelXbrl.modelDocument.xmlRootElement.iterchildren(f"{{http://www.xbrl.org/2003/linkbase}}{refType}Ref"),
            key=lambda elt: elt.get(refType + "URI"),
        ):
            dtsReferences.add(refElt.get("{http://www.w3.org/1999/xlink}href").partition("#")[0])
    dtsReferences = sorted(dtsReferences)  # turn into list
    footnoteFacts = set()

    def factFootnotes(
        fact: ModelFact,
        oimFact: dict[str, Any] | None = None,
        csvLinks: dict[str, Any] | None = None,
    ):
        footnotes = []
        oimLinks = {}
        if isCSVorXL and csvLinks is not None:
            oimLinks = csvLinks
        for footnoteRel in footnotesRelationshipSet.fromModelObject(fact):
            srcId = fact.id if fact.id else f"f{fact.objectIndex}"
            toObj = footnoteRel.toModelObject
            # json
            typePrefix = linkTypeAliases[footnoteRel.arcrole]
            groupPrefix = groupAliases[footnoteRel.linkrole]
            if typePrefix not in oimLinks:
                oimLinks[typePrefix] = {}
            _link = oimLinks[typePrefix]
            if groupPrefix not in _link:
                if isJSON:
                    _link[groupPrefix] = []
                elif isCSVorXL:
                    _link[groupPrefix] = {}
            tgtId = toObj.id if toObj.id else f"f{toObj.objectIndex}"
            if isJSON:
                tgtIdList = _link[groupPrefix]
                tgtIdList.append(tgtId)
            elif isCSVorXL:
                # Footnote links in xBRL-CSV include the CSV table identifier.
                tgtIdList = _link[groupPrefix].setdefault(f"facts.r_{srcId}.value", [])
                tgtIdList.append(f"footnotes.r_{tgtId}.footnote")
            footnote = {
                "group": footnoteRel.linkrole,
                "footnoteType": footnoteRel.arcrole,
            }
            if isinstance(toObj, ModelFact):
                footnote["factRef"] = tgtId
            else:  # text footnotes
                footnote["id"] = tgtId
                footnote["footnote"] = toObj.viewText()
                if toObj.xmlLang:
                    footnote["language"] = toObj.xmlLang
                footnoteFacts.add(toObj)
                footnotes.append(footnote)
        if oimLinks:
            _links = {
                typePrefix: {groupPrefix: idList for groupPrefix, idList in sorted(groups.items())}
                for typePrefix, groups in sorted(oimLinks.items())
            }
            if isJSON and oimFact is not None:
                oimFact["links"] = _links

        return footnotes

    def factAspects(fact: ModelFact) -> dict[str, Any]:
        oimFact = {}
        aspects = {}
        if isCSVorXL:
            oimFact["id"] = fact.id or f"f{fact.objectIndex}"
        concept = fact.concept
        if concept is not None:
            aspects[str(qnOimConceptAspect)] = oimValue(concept.qname)
            if concept.type is not None and concept.type.isOimTextFactType and fact.xmlLang:
                aspects[str(qnOimLangAspect)] = fact.xmlLang
        if fact.isItem:
            if fact.isNil:
                _value = None
            else:
                _inferredDecimals = inferredDecimals(fact)
                _value = oimValue(fact.xValue)
            oimFact["value"] = _value
            if fact.concept is not None and fact.concept.isNumeric:
                _numValue = fact.xValue
                if isinstance(_numValue, Decimal) and not isinf(_numValue) and not isnan(_numValue):
                    _numValue = int(_numValue) if _numValue == _numValue.to_integral() else float(_numValue)
                if not fact.isNil and not isinf(_inferredDecimals):  # accuracy omitted if infinite
                    oimFact["decimals"] = _inferredDecimals
        oimFact["dimensions"] = aspects
        cntx = fact.context
        if cntx is not None:
            if cntx.entityIdentifierElement is not None and cntx.entityIdentifier != ENTITY_NA_QNAME:
                aspects[str(qnOimEntityAspect)] = oimValue(qname(*cntx.entityIdentifier))
            if cntx.period is not None and not cntx.isForeverPeriod:
                aspects.update(oimPeriodValue(cntx))
            for _qn, dim in sorted(cntx.qnameDims.items(), key=lambda item: item[0]):
                if dim.isExplicit:
                    dimVal = oimValue(dim.memberQname)
                else:  # typed
                    if dim.typedMember.get("{http://www.w3.org/2001/XMLSchema-instance}nil") in ("true", "1"):
                        dimVal = None
                    else:
                        dimVal = dim.typedMember.stringValue
                aspects[str(dim.dimensionQname)] = dimVal
        unit = fact.unit
        if unit is not None:
            _mMul, _mDiv = unit.measures
            _sMul = "*".join(oimValue(m) for m in sorted(_mMul, key=lambda m: oimValue(m)))
            if _mDiv:
                _sDiv = "*".join(oimValue(m) for m in sorted(_mDiv, key=lambda m: oimValue(m)))
                if len(_mDiv) > 1:
                    _sUnit = f"({_sMul})/({_sDiv})" if len(_mMul) > 1 else f"{_sMul}/({_sDiv})"
                else:
                    _sUnit = f"({_sMul})/{_sDiv}" if len(_mMul) > 1 else f"{_sMul}/{_sDiv}"
            else:
                _sUnit = _sMul
            if _sUnit != "xbrli:pure":
                aspects[str(qnOimUnitAspect)] = _sUnit

        if isJSON:
            factFootnotes(fact, oimFact=oimFact)
        return oimFact

    # common metadata
    oimReport = {}  # top level of oim json output
    oimReport["documentInfo"] = oimDocInfo = {}
    oimDocInfo["documentType"] = nsOim + ("/xbrl-json" if isJSON else "/xbrl-csv")
    if isJSON:
        oimDocInfo["features"] = oimFeatures = {}
    oimDocInfo["namespaces"] = namespacePrefixes.namespaces
    if linkTypeAliases:
        oimDocInfo["linkTypes"] = {a: u for u, a in sorted(linkTypeAliases.items(), key=operator.itemgetter(1))}
    if groupAliases:
        oimDocInfo["linkGroups"] = {a: u for u, a in sorted(groupAliases.items(), key=operator.itemgetter(1))}
    oimDocInfo["taxonomy"] = dtsReferences
    if isJSON:
        oimFeatures["xbrl:canonicalValues"] = True

    factsToSave = modelXbrl.facts
    pluginData = modelXbrl.modelManager.cntlr.getPluginData(PLUGIN_NAME)
    if isinstance(pluginData, SaveLoadableOIMPluginData) and pluginData.deduplicateFactsType is not None:
        deduplicatedFacts = frozenset(ValidateDuplicateFacts.getDeduplicatedFacts(modelXbrl, pluginData.deduplicateFactsType))
        duplicateFacts = frozenset(f for f in modelXbrl.facts if f not in deduplicatedFacts)
        if duplicateFacts:
            for fact in duplicateFacts:
                ValidateDuplicateFacts.logDeduplicatedFact(modelXbrl, fact)
            factsToSave = [f for f in factsToSave if f not in duplicateFacts]

    if isJSON:
        # save JSON
        oimReport["facts"] = oimFacts = {}
        # add in report level extension objects
        if extensionReportObjects:
            for extObjQName, extObj in extensionReportObjects.items():
                oimReport[extObjQName] = extObj

        def saveJsonFacts(facts: list[ModelFact], oimFacts: dict[str, dict[str, Any]]) -> None:
            for fact in facts:
                oimFact = factAspects(fact)
                # add in fact level extension objects
                if extensionFactPropertiesMethod:
                    extensionFactPropertiesMethod(fact, oimFact)
                id = fact.id if fact.id else f"f{fact.objectIndex}"
                oimFacts[id] = oimFact
                if fact.modelTupleFacts:
                    saveJsonFacts(fact.modelTupleFacts, oimFacts)

        saveJsonFacts(factsToSave, oimFacts)

        # add footnotes as pseudo facts
        for ftObj in footnoteFacts:
            ftId = ftObj.id if ftObj.id else f"f{ftObj.objectIndex}"
            oimFacts[ftId] = oimFact = {}
            oimFact["value"] = ftObj.viewText()
            oimFact["dimensions"] = {
                "concept": "xbrl:note",
                "noteId": ftId,
            }
            if ftObj.xmlLang:
                oimFact["dimensions"]["language"] = ftObj.xmlLang.lower()

        # allow extension report final editing before writing json structure
        # (possible example, reorganize facts into array vs object)
        if extensionReportFinalizeMethod:
            extensionReportFinalizeMethod(oimReport)

        with io.StringIO() if outputZip else open(oimFile, "w", encoding="utf-8") as fh:
            fh.write(json.dumps(oimReport, indent=1))
            if outputZip:
                fh.seek(0)
                outputZip.writestr(os.path.basename(oimFile), fh.read())
        if not outputZip:
            modelXbrl.modelManager.cntlr.showStatus(_("Saved JSON OIM file {}").format(oimFile))

    elif isCSVorXL:
        # save CSV
        oimReport["tables"] = oimTables = {}
        oimReport["tableTemplates"] = csvTableTemplates = {}
        oimTables["facts"] = csvTable = {}
        csvTable["template"] = "facts"
        csvTable["url"] = "tbd"
        csvTableTemplates["facts"] = csvTableTemplate = {}
        csvTableTemplate["rowIdColumn"] = "id"
        csvTableTemplate["dimensions"] = csvTableDimensions = {}
        csvTableTemplate["columns"] = csvFactColumns = {}
        csvLinks = {}
        if footnotesRelationshipSet.modelRelationships:
            oimReport["links"] = csvLinks
        aspectQnCol = {}
        aspectsHeader = []

        def addAspectQnCol(aspectQn: QName | str) -> None:
            colQName = str(aspectQn).replace("xbrl:", "")
            colNCName = colQName.replace(":", "_")
            if colNCName not in aspectQnCol:
                aspectQnCol[colNCName] = len(aspectsHeader)
                aspectsHeader.append(colNCName)
                if colNCName == "value":
                    csvFactColumns[colNCName] = col = {}
                    col["dimensions"] = {}
                elif colNCName == "id":
                    csvFactColumns[colNCName] = {}  # empty object
                elif colNCName == "decimals":
                    csvFactColumns[colNCName] = {}  # empty object
                    csvTableTemplate[colQName] = "$" + colNCName
                else:
                    csvFactColumns[colNCName] = {}  # empty object
                    csvTableDimensions[colQName] = "$" + colNCName

        # pre-ordered aspect columns
        # if hasId:
        addAspectQnCol("id")
        addAspectQnCol(qnOimConceptAspect)
        if hasNumeric:
            addAspectQnCol("decimals")
        if qnOimEntityAspect in aspectsDefined:
            addAspectQnCol(qnOimEntityAspect)
        if qnOimPeriodAspect in aspectsDefined:
            addAspectQnCol(qnOimPeriodAspect)
        if qnOimUnitAspect in aspectsDefined:
            addAspectQnCol(qnOimUnitAspect)
        for aspectQn in sorted(aspectsDefined, key=lambda qn: str(qn)):
            if aspectQn.namespaceURI != nsOim:
                addAspectQnCol(aspectQn)
        addAspectQnCol("value")

        def aspectCols(fact: ModelFact) -> list[Any]:
            cols: list[Any] = [None] * len(aspectsHeader)

            def setColValues(aspects: dict[str, Any]) -> None:
                for aspectQn, aspectValue in aspects.items():
                    colQName = str(aspectQn).replace("xbrl:", "")
                    colNCName = colQName.replace(":", "_")
                    if isinstance(aspectValue, dict):
                        setColValues(aspectValue)
                    elif colNCName in aspectQnCol:
                        if aspectValue is None:
                            _aspectValue = "#nil"
                        elif aspectValue == "":
                            _aspectValue = "#empty"
                        elif isinstance(aspectValue, str) and aspectValue.startswith("#"):
                            _aspectValue = "#" + aspectValue
                        else:
                            _aspectValue = aspectValue
                        cols[aspectQnCol[colNCName]] = _aspectValue

            setColValues(factAspects(fact))
            return cols

        # metadata

        if isCSV:
            if oimFile.endswith("-facts.csv"):  # strip -facts.csv if a prior -facts.csv file was chosen
                _baseURL = oimFile[:-10]
            elif oimFile.endswith(".csv"):
                _baseURL = oimFile[:-4]
            else:
                _baseURL = oimFile
            _csvInfo = {}  # open file, writer

            def _open(filesuffix: str | None, tabname: str, csvTable: dict[str, Any] | None = None) -> None:
                _filename = _baseURL if filesuffix is None else _baseURL + filesuffix
                if csvTable is not None:
                    csvTable["url"] = os.path.basename(_filename)  # located in same directory with metadata
                _csvInfo["file"] = open(_filename, csvOpenMode, newline=csvOpenNewline, encoding="utf-8-sig")
                _csvInfo["writer"] = csv.writer(_csvInfo["file"], dialect="excel")

            def _writerow(row: Iterable[str], header: bool = False) -> None:
                _csvInfo["writer"].writerow(row)

            def _close() -> None:
                _csvInfo["file"].close()
                _csvInfo.clear()
        elif isXL:
            headerWidths = {
                "concept": 40,
                "decimals": 8,
                "language": 9,
                "value": 50,
                "entity": 20,
                "period": 20,
                "unit": 20,
                "metadata": 100,
            }

            # Excel's light orange fill color = 00FF990
            hdrCellFill = PatternFill(patternType=fills.FILL_SOLID, fgColor=Color("00FFBF5F"))
            workbook = Workbook()
            # remove pre-existing worksheets
            while len(workbook.worksheets) > 0:
                workbook.remove(workbook.worksheets[0])
            _xlInfo = {}  # open file, writer

            def _open(filesuffix: str | None, tabname: str, csvTable: dict[str, Any] | None = None) -> None:
                if csvTable is not None:
                    csvTable["url"] = tabname + "!"
                _xlInfo["ws"] = workbook.create_sheet(title=tabname)

            def _writerow(rowvalues: Iterable[str], header: bool = False) -> None:
                row = []
                _ws = _xlInfo["ws"]
                for i, v in enumerate(rowvalues):
                    cell = WriteOnlyCell(_ws, value=v)
                    if header:
                        cell.fill = hdrCellFill
                        cell.alignment = Alignment(horizontal="center", vertical="center", wrap_text=True)
                        colLetter = chr(ord("A") + i)
                        _ws.column_dimensions[colLetter] = ColumnDimension(_ws, customWidth=True)
                        _ws.column_dimensions[colLetter].width = headerWidths.get(v, 40)

                    else:
                        cell.alignment = Alignment(
                            horizontal="right"
                            if isinstance(v, Number)
                            else "center"
                            if isinstance(v, bool)
                            else "left",
                            vertical="top",
                            wrap_text=isinstance(v, str),
                        )
                    row.append(cell)
                _ws.append(row)

            def _close() -> None:
                _xlInfo.clear()

        # save facts
        _open("-facts.csv", "facts", csvTable)
        _writerow(aspectsHeader, header=True)

        def saveCSVfacts(facts: list[ModelFact]) -> None:
            for fact in facts:
                _writerow(aspectCols(fact))
                saveCSVfacts(fact.modelTupleFacts)

        saveCSVfacts(factsToSave)
        _close()

        # save footnotes
        if footnotesRelationshipSet.modelRelationships:
            footnotesDeduplicatedById = {
                footnote["id"]: footnote
                for fact in modelXbrl.facts
                for footnote in factFootnotes(fact, csvLinks=csvLinks)
            }
            footnotes = sorted(footnotesDeduplicatedById.values(), key=operator.itemgetter("id"))
            if footnotes:  # text footnotes
                footnotesTable = {"template": "footnotes"}
                oimTables["footnotes"] = footnotesTable
                csvTableTemplates[footnotesTable["template"]] = {
                    "rowIdColumn" : "id",
                    "dimensions": {
                        "language": "$language"
                    },
                    "columns": {
                        "id": {},
                        "footnote": {
                            "dimensions": {
                                "concept": "xbrl:note",
                            },
                        },
                        "language": {},
                    }
                }
                _open("-footnotes.csv", "footnotes", footnotesTable)
                cols = ("id", "footnote", "language")
                _writerow(cols, header=True)
                for footnote in footnotes:
                    _writerow(tuple(footnote.get(col, "") for col in cols))
                _close()

        # save metadata
        if isCSV:
            assert isinstance(_baseURL, str)
            csvMetadataFile = _baseURL + "-metadata.json"
            with open(csvMetadataFile, "w", encoding="utf-8") as fh:
                fh.write(json.dumps(oimReport, ensure_ascii=False, indent=2, sort_keys=False))
            modelXbrl.modelManager.cntlr.showStatus(_("Saved CSV OIM metadata file {}").format(csvMetadataFile))
        elif isXL:
            _open(None, "metadata")
            _writerow(["metadata"], header=True)
            _writerow([json.dumps(oimReport, ensure_ascii=False, indent=1, sort_keys=False)])
            _close()
            workbook.save(oimFile)
            modelXbrl.modelManager.cntlr.showStatus(_("Saved Excel file {}").format(oimFile))


def saveLoadableOIMMenuCommand(cntlr: CntlrWinMain) -> None:
    # save DTS menu item has been invoked
    if (
        cntlr.config is None
        or cntlr.modelManager is None
        or cntlr.modelManager.modelXbrl is None
        or cntlr.modelManager.modelXbrl.modelDocument is None
        or cntlr.modelManager.modelXbrl.modelDocument.type
        not in (ModelDocumentType.INSTANCE, ModelDocumentType.INLINEXBRL, ModelDocumentType.INLINEXBRLDOCUMENTSET)
    ):
        cntlr.addToLog(
            messageCode="arelleOIMsaver",
            message=_("No supported XBRL instance documents loaded that can be saved to OIM format."),
        )
        return
        # get file name into which to save log file while in foreground thread
    oimFile = cntlr.uiFileDialog(
        "save",
        title=_("arelle - Save Loadable OIM file"),
        initialdir=cntlr.config.setdefault("loadableOIMFileDir", "."),
        filetypes=[(_("JSON file .json"), "*.json"), (_("CSV file .csv"), "*.csv"), (_("XLSX file .xlsx"), "*.xlsx")],
        defaultextension=".json",
    )  # type: ignore[no-untyped-call]
    if not isinstance(oimFile, str):
        # User cancelled file dialog.
        return

    cntlr.config["loadableOIMFileDir"] = os.path.dirname(oimFile)
    cntlr.saveConfig()

    thread = threading.Thread(
        target=lambda _modelXbrl=cntlr.modelManager.modelXbrl, _oimFile=oimFile: saveLoadableOIM(
            _modelXbrl, _oimFile, None
        )
    )
    thread.daemon = True
    thread.start()


def saveOimFiles(
    cntlr: Cntlr,
    modelXbrl: ModelXbrl,
    oimFiles: list[str],
    responseZipStream: BinaryIO | None = None,
) -> None:
    try:
        if responseZipStream is None:
            for oimFile in oimFiles:
                saveLoadableOIM(modelXbrl, oimFile, None)
        else:
            with zipfile.ZipFile(responseZipStream, "a", zipfile.ZIP_DEFLATED, True) as _zip:
                for oimFile in oimFiles:
                    saveLoadableOIM(modelXbrl, oimFile, _zip)
            responseZipStream.seek(0)
    except Exception as ex:
        cntlr.addToLog(f"Exception saving OIM {ex}")


@dataclass
class SaveLoadableOIMPluginData(PluginData):
    deduplicateFactsType: ValidateDuplicateFacts.DeduplicationType | None
    saveTestcaseOimFileSuffix: str | None


class SaveLoadableOIMPlugin(PluginHooks):
    @staticmethod
    def cntlrCmdLineOptions(
        parser: OptionParser,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        parser.add_option(
            "--saveLoadableOIM",
            action="store",
            dest="saveLoadableOIM",
            help=_("Save Loadable OIM file (JSON, CSV or XLSX)"),
        )
        parser.add_option(
            "--saveLoadableOIMDirectory",
            action="store",
            dest="saveLoadableOIMDirectory",
            help=_("Directory to export all supported OIM formats (JSON and CSV)"),
        )
        parser.add_option(
            "--saveTestcaseOIM",
            action="store",
            dest="saveTestcaseOimFileSuffix",
            help=_("Save Testcase Variation OIM file (argument file suffix and type, such as -savedOim.csv"),
        )
        parser.add_option(
            "--deduplicateOimFacts",
            action="store",
            choices=[a.value for a in ValidateDuplicateFacts.DeduplicationType],
            dest="deduplicateOimFacts",
            help=_("Remove duplicate facts when saving the OIM instance"))

    @staticmethod
    def cntlrCmdLineUtilityRun(
        cntlr: Cntlr,
        options: RuntimeOptions,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        deduplicateOimFacts = cast(Optional[str], getattr(options, "deduplicateOimFacts", None))
        saveTestcaseOimFileSuffix = cast(Optional[str], getattr(options, "saveTestcaseOimFileSuffix", None))
        deduplicateFactsType = None
        if deduplicateOimFacts is not None:
            deduplicateFactsType = ValidateDuplicateFacts.DeduplicationType(deduplicateOimFacts)
        pluginData = SaveLoadableOIMPluginData(PLUGIN_NAME, deduplicateFactsType, saveTestcaseOimFileSuffix)
        cntlr.setPluginData(pluginData)

    @staticmethod
    def cntlrWinMainMenuTools(
        cntlr: CntlrWinMain,
        menu: Menu,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        menu.add_command(label="Save Loadable OIM", underline=0, command=lambda: saveLoadableOIMMenuCommand(cntlr))

    @staticmethod
    def cntlrCmdLineXbrlRun(
        cntlr: CntlrCmdLine,
        options: RuntimeOptions,
        modelXbrl: ModelXbrl,
        entrypoint: dict[str, str] | None = None,
        sourceZipStream: BinaryIO | None = None,
        responseZipStream: BinaryIO | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        # extend XBRL-loaded run processing for this option
        oimFile = cast(Optional[str], getattr(options, "saveLoadableOIM", None))
        allOimDirectory = cast(Optional[str], getattr(options, "saveLoadableOIMDirectory", None))
        if (oimFile or allOimDirectory) and (
            modelXbrl is None
            or modelXbrl.modelDocument is None
            or modelXbrl.modelDocument.type
            not in {
                ModelDocumentType.INSTANCE,
                ModelDocumentType.INLINEXBRL,
                ModelDocumentType.INLINEXBRLDOCUMENTSET,
            }
        ):
            cntlr.addToLog("No XBRL instance has been loaded.")
            return
        if oimFile:
            saveOimFiles(cntlr, modelXbrl, [oimFile], responseZipStream)
        if allOimDirectory:
            oimDir = Path(allOimDirectory)
            try:
                oimDir.mkdir(parents=True, exist_ok=True)
            except OSError as err:
                cntlr.addToLog(
                    _("Unable to save OIM files into requested directory: {}, {}").format(allOimDirectory, err.strerror)
                )
                return
            assert modelXbrl.modelDocument is not None
            basefileStem = None
            if hasattr(modelXbrl, "ixdsTarget"):
                ixdsTarget = modelXbrl.ixdsTarget
                for doc in modelXbrl.modelDocument.referencesDocument:
                    if doc.type in {ModelDocumentType.INLINEXBRL, ModelDocumentType.INSTANCE}:
                        instanceFilename = Path(doc.uri)
                        if ixdsTarget is not None:
                            basefileStem = f"{instanceFilename.stem}.{ixdsTarget}"
                        else:
                            basefileStem = instanceFilename.stem
                        break
            if basefileStem is None:
                basefileStem = Path(modelXbrl.modelDocument.basename).stem
            oimFiles = [str(oimDir.joinpath(basefileStem + ext)) for ext in (".csv", ".json")]
            saveOimFiles(cntlr, modelXbrl, oimFiles, responseZipStream)

    @staticmethod
    def testcaseVariationValidated(
        testcaseDTS: ModelXbrl,
        testInstanceDTS: ModelXbrl,
        extraErrors: list[str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        pluginData = testcaseDTS.modelManager.cntlr.getPluginData(PLUGIN_NAME)
        oimFileSuffix = (
            pluginData.saveTestcaseOimFileSuffix if isinstance(pluginData, SaveLoadableOIMPluginData) else None
        )
        if (
            oimFileSuffix
            and testInstanceDTS is not None
            and testInstanceDTS.modelDocument is not None
            and testInstanceDTS.modelDocument.type
            in (ModelDocumentType.INSTANCE, ModelDocumentType.INLINEXBRL, ModelDocumentType.INLINEXBRLDOCUMENTSET)
            and not any(oimErrorPattern.match(error) for error in testInstanceDTS.errors if error is not None)
        ):  # no OIM errors
            try:
                saveLoadableOIM(testInstanceDTS, testInstanceDTS.modelDocument.uri + oimFileSuffix)
            except Exception as ex:
                testcaseDTS.modelManager.cntlr.addToLog(f"Exception saving OIM {ex}")

    @staticmethod
    def saveLoadableOimSave(
        modelXbrl: ModelXbrl,
        oimFile: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        return saveLoadableOIM(modelXbrl, oimFile, *args, **kwargs)


__pluginInfo__ = {
    "name": PLUGIN_NAME,
    "version": "1.3",
    "description": "This plug-in saves XBRL in OIM JSON, CSV or XLSX that can be re-loaded per se.",
    "license": "Apache-2",
    "author": authorLabel,
    "copyright": copyrightLabel,
    # classes of mount points (required)
    "CntlrWinMain.Menu.Tools": SaveLoadableOIMPlugin.cntlrWinMainMenuTools,
    "CntlrCmdLine.Options": SaveLoadableOIMPlugin.cntlrCmdLineOptions,
    "CntlrCmdLine.Utility.Run": SaveLoadableOIMPlugin.cntlrCmdLineUtilityRun,
    "CntlrCmdLine.Xbrl.Run": SaveLoadableOIMPlugin.cntlrCmdLineXbrlRun,
    "TestcaseVariation.Validated": SaveLoadableOIMPlugin.testcaseVariationValidated,
    "SaveLoadableOim.Save": SaveLoadableOIMPlugin.saveLoadableOimSave,
}



'''
See COPYRIGHT.md for copyright information.
'''
import decimal, os
from collections import defaultdict

from arelle import LinkRelationships
from arelle.ModelDocumentType import ModelDocumentType
from arelle import XbrlConst
from arelle.typing import LocPrototypeBase, PrototypeElementTreeBase, PrototypeObjectBase
import arelle.XmlValidate
ModelDocument = None

class PrototypeObject(PrototypeObjectBase):
    def __init__(self, modelDocument, sourceElement=None):
        self.modelDocument = modelDocument
        self.sourceElement = sourceElement
        self.attributes = {}

    @property
    def sourceline(self):
        return self.sourceElement.sourceline if self.sourceElement is not None else None

    def get(self, key, default=None):
        return self.attributes.get(key, default)

    def itersiblings(self, **kwargs):
        """Method proxy for itersiblings() of lxml arc element"""
        return self.sourceElement.itersiblings(**kwargs) if self.sourceElement is not None else ()

    def getparent(self):
        """(_ElementBase) -- Method proxy for getparent() of lxml arc element"""
        return self.sourceElement.getparent() if self.sourceElement is not None else None

    def iterchildren(self):
        yield from () # no children

    def iterdescendants(self):
        for elt in self.iterchildren():
            yield elt
            for e in elt.iterdescendants():
                yield e

class LinkPrototype(PrototypeObject, LinkRelationships.LinkRelationships):      # behaves like a ModelLink for relationship prototyping
    def __init__(self, modelDocument, parent, qname, role, sourceElement=None):
        super(LinkPrototype, self).__init__(modelDocument, sourceElement)
        self._parent = parent
        self.modelXbrl = modelDocument.modelXbrl
        self.qname = self.elementQname = qname
        self.namespaceURI = qname.namespaceURI
        self.localName = qname.localName
        self.role = role
        # children are arc and loc elements or prototypes
        self.childElements = []
        self.text = self.textValue = None
        self.attributes = {"{http://www.w3.org/1999/xlink}type":"extended"}
        if role:
            self.attributes["{http://www.w3.org/1999/xlink}role"] = role
        self.labeledResources = defaultdict(list)
        self.initRelationships()

    def clear(self):
        self.__dict__.clear() # dereference here, not an lxml object, don't use superclass clear()

    def __iter__(self):
        return iter(self.childElements)

    def getparent(self):
        return self._parent

    def iterchildren(self):
        return iter(self.childElements)

class LocPrototype(PrototypeObject, LocPrototypeBase):
    def __init__(self, modelDocument, parent, label, locObject, role=None, sourceElement=None):
        super(LocPrototype, self).__init__(modelDocument, sourceElement)
        self._parent = parent
        self.modelXbrl = modelDocument.modelXbrl
        self.qname = self.elementQname = XbrlConst.qnLinkLoc
        self.namespaceURI = self.qname.namespaceURI
        self.localName = self.qname.localName
        self.text = self.textValue = None
        # children are arc and loc elements or prototypes
        self.attributes = {"{http://www.w3.org/1999/xlink}type":"locator",
                           "{http://www.w3.org/1999/xlink}label":label}
        # add an href if it is a 1.1 id
        if isinstance(locObject,str): # it is an id
            self.attributes["{http://www.w3.org/1999/xlink}href"] = "#" + locObject
        if role:
            self.attributes["{http://www.w3.org/1999/xlink}role"] = role
        self.locObject = locObject

    def clear(self):
        self.__dict__.clear() # dereference here, not an lxml object, don't use superclass clear()

    @property
    def xlinkLabel(self):
        return self.attributes.get("{http://www.w3.org/1999/xlink}label")

    def dereference(self):
        if isinstance(self.locObject,str): # dereference by ID
            return self.modelDocument.idObjects.get(self.locObject,None) # id may not exist
        else: # it's an object pointer
            return self.locObject

    def getparent(self):
        return self._parent

    def get(self, key, default=None):
        return self.attributes.get(key, default)

class ArcPrototype(PrototypeObject):
    def __init__(self, modelDocument, parent, qname, fromLabel, toLabel, linkrole, arcrole, order="1", sourceElement=None):
        super(ArcPrototype, self).__init__(modelDocument, sourceElement)
        self._parent = parent
        self.modelXbrl = modelDocument.modelXbrl
        self.qname = self.elementQname = qname
        self.namespaceURI = qname.namespaceURI
        self.localName = qname.localName
        self.linkrole = linkrole
        self.arcrole = arcrole
        self.order = order
        self.text = self.textValue = None
        # children are arc and loc elements or prototypes
        self.attributes = {"{http://www.w3.org/1999/xlink}type":"arc",
                           "{http://www.w3.org/1999/xlink}from": fromLabel,
                           "{http://www.w3.org/1999/xlink}to": toLabel,
                           "{http://www.w3.org/1999/xlink}arcrole": arcrole}
        # must look validated (because it can't really be validated)
        self.xValid = arelle.XmlValidate.VALID
        self.xValue = self.sValue = None
        self.xAttributes = {}

    @property
    def orderDecimal(self):
        return decimal.Decimal(self.order)

    def clear(self):
        self.__dict__.clear() # dereference here, not an lxml object, don't use superclass clear()

    @property
    def arcElement(self):
        return self.sourceElement if self.sourceElement is not None else None

    def getparent(self):
        return self._parent

    def get(self, key, default=None):
        return self.attributes.get(key, default)

    def items(self):
        return self.attributes.items()

class DocumentPrototype():
    def __init__(self, modelXbrl, uri, base=None, referringElement=None, isEntry=False, isDiscovered=False, isIncluded=None, namespace=None, reloadCache=False, **kwargs):
        self.modelXbrl = modelXbrl
        self.skipDTS = modelXbrl.skipDTS
        self.modelDocument = self
        if referringElement is not None:
            if referringElement.localName == "schemaRef":
                self.type = ModelDocumentType.SCHEMA
            elif referringElement.localName == "linkbaseRef":
                self.type = ModelDocumentType.LINKBASE
            else:
                self.type = ModelDocumentType.UnknownXML
        else:
            self.type = ModelDocumentType.UnknownXML
        normalizedUri = modelXbrl.modelManager.cntlr.webCache.normalizeUrl(uri, base)
        self.filepath = modelXbrl.modelManager.cntlr.webCache.getfilename(normalizedUri, filenameOnly=True)
        self.uri = modelXbrl.modelManager.cntlr.webCache.normalizeUrl(self.filepath)
        self.basename = os.path.basename(self.filepath)
        self.targetNamespace = None
        self.referencesDocument = {}
        self.hrefObjects = []
        self.schemaLocationElements = set()
        self.referencedNamespaces = set()
        self.inDTS = False
        self.xmlRootElement = None


    def clear(self):
        self.__dict__.clear() # dereference here, not an lxml object, don't use superclass clear()

class PrototypeElementTree(PrototypeElementTreeBase): # equivalent to _ElementTree for parenting root element in non-lxml situations
    def __init__(self, rootElement):
        self.rootElement = rootElement

    def getroot(self):
        return self.rootElement

    def iter(self):
        yield self.rootElement
        for e in self.rootElement.iterdescendants():
            yield e

    def ixIter(self, childOnly=False):
        yield self.rootElement
        if not childOnly:
            for e in self.rootElement.ixIter(childOnly):
                yield e



'''
See COPYRIGHT.md for copyright information.
'''
import os
from datetime import timedelta
from collections import OrderedDict
from copy import deepcopy
from lxml import etree
from arelle import ViewFile
from arelle.FunctionXs import xsString
from arelle.ModelObject import ModelObject
from arelle.Aspect import Aspect, aspectModels, aspectRuleAspects, aspectModelAspect, aspectStr
from arelle.rendering.RenderingLayout import layoutTable
from arelle import XbrlConst
from arelle.XmlUtil import elementFragmentIdentifier, addQnameValue

def viewRenderedLayout(modelXbrl, outfile, lang=None, viewTblELR=None, sourceView=None, diffToFile=False, cssExtras=""):
    modelXbrl.modelManager.showStatus(_("saving rendering"))
    view = ViewRenderedLayout(modelXbrl, outfile, lang, cssExtras)

    if sourceView is not None:
        if sourceView.lytMdlTblMdl:
            lytMdlTblMdl = sourceView.lytMdlTblMdl
        else:
            lytMdlTblMdl = layoutTable(sourceView)
    else:
        layoutTable(view)
        lytMdlTblMdl = view.lytMdlTblMdl
    if view.tblElt is not None: # may be None if there is no table
        view.view(lytMdlTblMdl)
    if diffToFile and outfile:
        from arelle.ValidateInfoset import validateRenderingInfoset
        validateRenderingInfoset(modelXbrl, outfile, view.xmlDoc)
        view.close(noWrite=True)
    else:
        view.close()
    modelXbrl.modelManager.showStatus(_("rendering saved to {0}").format(outfile), clearAfter=5000)

class ViewRenderedLayout(ViewFile.View):
    def __init__(self, modelXbrl, outfile, lang, cssExtras):
        # find table model namespace based on table namespace
        self.tableModelNamespace = XbrlConst.tableModel
        for xsdNs in modelXbrl.namespaceDocs.keys():
            if xsdNs in (XbrlConst.tableMMDD, XbrlConst.table):
                self.tableModelNamespace = xsdNs + "/model"
                break
        super(ViewRenderedLayout, self).__init__(modelXbrl, outfile,
                                               f'tableModel xmlns="{self.tableModelNamespace}" xmlns:xbrli="http://www.xbrl.org/2003/instance"',
                                               lang,
                                               style="rendering",
                                               cssExtras=cssExtras)
        class nonTkBooleanVar():
            def __init__(self, value=True):
                self.value = value
            def set(self, value):
                self.value = value
            def get(self):
                return self.value
        # context menu boolean vars (non-tkinter boolean
        self.ignoreDimValidity = nonTkBooleanVar(value=True)


    def tableModelQName(self, localName):
        return '{' + self.tableModelNamespace + '}' + localName

    def viewReloadDueToMenuAction(self, *args):
        self.view()

    def view(self, lytMdlTblMdl):

        self.tblElt.append(etree.Comment("Entry point file: {0}".format(lytMdlTblMdl.entryPointUrl)))

        for lytMdlTableSet in lytMdlTblMdl.lytMdlTableSets:
            tableSetElt = etree.SubElement(self.tblElt, self.tableModelQName("tableSet"))
            tableSetElt.append(etree.Comment(f"TableSet linkbase file: {lytMdlTableSet.srcFile}, line {lytMdlTableSet.srcLine}"))
            tableSetElt.append(etree.Comment(f"TableSet linkrole: {lytMdlTableSet.srcLinkrole}"))
            etree.SubElement(tableSetElt, self.tableModelQName("label")).text = lytMdlTableSet.label
            for lytMdlTable in lytMdlTableSet.lytMdlTables:
                tableElt = etree.SubElement(tableSetElt, self.tableModelQName("table"))
                for name, value in lytMdlTable.strctMdlTable.tblParamValues.items():
                    tableElt.append(etree.Comment(f" ${name} = \"{value}\" "))
                for lytMdlHeaders in lytMdlTable.lytMdlHeaders:
                    hdrsElt = etree.SubElement(tableElt, self.tableModelQName("headers"), attrib={"axis": lytMdlHeaders.axis})
                    for lytMdlGroup in lytMdlHeaders.lytMdlGroups:
                        groupElt = etree.SubElement(hdrsElt, self.tableModelQName("group"))
                        groupElt.append(etree.Comment(f"Breakdown node file: {lytMdlGroup.srcFile}, line {lytMdlGroup.srcLine}"))
                        if lytMdlGroup.label:
                            etree.SubElement(groupElt, self.tableModelQName("label")).text=lytMdlGroup.label
                        for lytMdlHeader in lytMdlGroup.lytMdlHeaders:
                            if all(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlHeader.lytMdlCells):
                                continue # skip header with only open aspect entry surrogates
                            hdrElt = etree.SubElement(groupElt, self.tableModelQName("header"))
                            for lytMdlCell in lytMdlHeader.lytMdlCells:
                                if lytMdlCell.isOpenAspectEntrySurrogate:
                                    continue # strip all open aspect entry surrogates from layout model file
                                attrib = {}
                                if lytMdlCell.span > 1:
                                    attrib["span"] = str(lytMdlCell.span)
                                if lytMdlCell.rollup:
                                    attrib["rollup"] = "true"
                                cellElt = etree.SubElement(hdrElt, self.tableModelQName("cell"), attrib)
                                if lytMdlCell.id:
                                    cellElt.append(etree.Comment(f"Cell id {lytMdlCell.id}"))
                                for label, role, lang in lytMdlCell.labels:
                                    if role or lang:
                                        cellElt.append(etree.Comment(f"Label role: {role}, lang {lang}"))
                                    labelElt = etree.SubElement(cellElt, self.tableModelQName("label")).text = label
                                for lytMdlConstraint in lytMdlCell.lytMdlConstraints:
                                    attrib = None
                                    if lytMdlConstraint.tag:
                                        attrib = {"tag": lytMdlConstraint.tag}
                                    constraintElt = etree.SubElement(cellElt, self.tableModelQName("constraint"), attrib)
                                    etree.SubElement(constraintElt, self.tableModelQName("aspect")).text=aspectStr(lytMdlConstraint.aspect)
                                    valueElt = etree.SubElement(constraintElt, self.tableModelQName("value"))
                                    aspect = lytMdlConstraint.aspect
                                    aspectValue = lytMdlConstraint.value
                                    if aspect == Aspect.PERIOD:
                                        if isinstance(aspectValue, ModelObject):
                                            for perElt in aspectValue.iterchildren():
                                                valueElt.append(deepcopy(perElt))
                                        elif isinstance(aspectValue, dict) and "periodType" in aspectValue:
                                            if aspectValue["periodType"] == "duration":
                                                aspectElt = etree.Element("{http://www.xbrl.org/2003/instance}startDate")
                                                aspectElt.text = f"{aspectValue['startDate'].strftime('%Y-%m-%dT%H:%M:%S.%f')}Z"
                                                valueElt.append(aspectElt)

                                                aspectElt = etree.Element("{http://www.xbrl.org/2003/instance}endDate")
                                                aspectElt.text = f"{(aspectValue['endDate'] + timedelta(days=1)).strftime('%Y-%m-%dT%H:%M:%S.%f')}Z"
                                                valueElt.append(aspectElt)
                                            elif aspectValue["periodType"] == "instant":
                                                aspectElt = etree.Element("{http://www.xbrl.org/2003/instance}instant")
                                                aspectElt.text = f"{(aspectValue['instant'] + timedelta(days=1)).strftime('%Y-%m-%dT%H:%M:%S.%f')}Z"
                                                valueElt.append(aspectElt)
                                            elif aspectValue["periodType"] == "forever":
                                                aspectElt = etree.Element("{http://www.xbrl.org/2003/instance}forever")
                                                valueElt.append(aspectElt)
                                    elif aspect == Aspect.ENTITY_IDENTIFIER:
                                        if isinstance(aspectValue, ModelObject):
                                            valueElt.append(deepcopy(aspectValue))
                                        elif isinstance(aspectValue, dict):
                                            aspectElt = etree.Element("{http://www.xbrl.org/2003/instance}identifier", {"scheme": aspectValue['scheme']})
                                            aspectElt.text = str(aspectValue['identifier'])
                                            valueElt.append(aspectElt)
                                    elif aspect == Aspect.UNIT:
                                        aspectElt = etree.SubElement(valueElt, "{http://www.xbrl.org/2003/instance}unit")
                                        if isinstance(aspectValue, ModelObject):
                                            for unitElt in aspectValue.iterchildren():
                                                aspectElt.append(deepcopy(unitElt))
                                        elif isinstance(aspectValue, dict):
                                            for m in aspectValue['measures']:
                                                etree.SubElement(aspectElt, "{http://www.xbrl.org/2003/instance}measure").text = str(m)
                                    else:
                                        if isinstance(aspectValue, ModelObject):
                                            valueElt.append(deepcopy(aspectValue))
                                        elif not (aspectValue is None and aspect in self.modelXbrl.qnameConcepts and self.modelXbrl.qnameConcepts[aspect].isExplicitDimension):
                                            # don't put None for value of explicit dimension which id defaulted
                                            valueElt.text = xsString(None, None, addQnameValue(self.xmlDoc, aspectValue))
                for lytMdlZCell in lytMdlTable.lytMdlBodyChildren:
                    zCellsElt = etree.SubElement(tableElt, self.tableModelQName("cells"), attrib={"axis": "z"})
                    for lytMdlYCell in lytMdlZCell.lytMdlBodyChildren:
                        yCellsElt = etree.SubElement(zCellsElt, self.tableModelQName("cells"), attrib={"axis": "y"})
                        for lytMdlXCell in lytMdlYCell.lytMdlBodyChildren:
                            if not any(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlXCell.lytMdlBodyChildren):
                                xCellsElt = etree.SubElement(yCellsElt, self.tableModelQName("cells"), attrib={"axis": "x"})
                                for lytMdlCell in lytMdlXCell.lytMdlBodyChildren:
                                    if lytMdlCell.isOpenAspectEntrySurrogate:
                                        continue
                                    cellElt = etree.SubElement(xCellsElt, self.tableModelQName("cell"))
                                    for f, v, justify in lytMdlCell.facts:
                                        cellElt.append(etree.Comment(
                                            f"{f.qname}: context {f.contextID}, value {v[:32]}, file {f.modelDocument.basename}, line {f.sourceline}"))
                                        if v is not None:
                                            etree.SubElement(cellElt, self.tableModelQName("fact")
                                                ).text = f"{f.modelDocument.basename}#{elementFragmentIdentifier(f)}"



"""
See COPYRIGHT.md for copyright information.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from arelle import PluginManager, PackageManager
from arelle.packages.report.ReportPackageValidator import ReportPackageValidator

if TYPE_CHECKING:
    from arelle.Cntlr import Cntlr
    from arelle.FileSource import FileSource


class ValidateFileSource:
    def __init__(self, cntrl: Cntlr, filesource: FileSource):
        self._cntrl = cntrl
        self._filesource = filesource

    def validate(self, forceValidateAsReportPackages: bool = False, forceValidateAsTaxonomyPackage: bool = False, errors: list[str] | None = None) -> None:
        for pluginXbrlMethod in PluginManager.pluginClassMethods("Validate.FileSource"):
            pluginXbrlMethod(self._cntrl, self._filesource)
        if self._filesource.isReportPackage or forceValidateAsReportPackages:
            rpValidator = ReportPackageValidator(self._filesource)
            for val in rpValidator.validate():
                codes = [val.codes] if isinstance(val.codes, str) else val.codes
                for code in codes:
                    self._cntrl.error(
                        level=val.level.name,
                        codes=code,
                        msg=val.msg,
                        fileSource=self._filesource,
                        **val.args
                    )
                    if errors is not None:
                        errors.append(code)
        # Automatically validates a report package as a taxonomy package if a
        # taxonomy package metadata file exists.
        self._filesource.loadTaxonomyPackageMappings(
            errors=errors,
            expectTaxonomyPackage=forceValidateAsTaxonomyPackage
        )
